@@ -1,16 +1,30 @@
 """ Functions to find optimal index parameters """
-
+import json
+import logging
+import math
+import os
 import re
+import tempfile
 from functools import partial, reduce
 from math import floor, log2, sqrt
 from operator import mul
-from typing import Callable, List, Optional, TypeVar
+from typing import Callable, List, Optional, TypeVar, Dict, Union
 
 import faiss
-from autofaiss.external.metadata import IndexMetadata
-from autofaiss.indices.index_utils import set_search_hyperparameters, speed_test_ms_per_query
+import fsspec
+
+from autofaiss.external.metadata import compute_memory_necessary_for_training_wrapper, IndexMetadata
+from autofaiss.external.scores import compute_fast_metrics
+from autofaiss.indices.index_utils import (
+    set_search_hyperparameters,
+    speed_test_ms_per_query,
+    parallel_download_indices_from_remote,
+)
 from autofaiss.utils.algorithms import discrete_binary_search
 from autofaiss.utils.cast import cast_memory_to_bytes
+from autofaiss.utils.decorators import Timeit
+
+logger = logging.getLogger("autofaiss")
 
 
 def check_if_index_needs_training(index_key: str) -> bool:
@@ -48,21 +62,8 @@ def index_key_to_nb_cluster(index_key: str) -> int:
     return nb_clusters
 
 
-def compute_memory_necessary_for_training(nb_training_vectors: int, index_key: str, vec_dim: int) -> float:
-    """
-    Function that computes the memory necessary to train an index with nb_training_vectors vectors
-    """
-
-    memory_per_vector = 4.0 * vec_dim
-    matching_opq = re.search(r"OPQ\d+_(\d+)", index_key)
-    if matching_opq:
-        input_size_opq = int(matching_opq.group(1))
-        memory_per_vector += input_size_opq * 4.0
-    return memory_per_vector * nb_training_vectors
-
-
 def get_optimal_train_size(
-    nb_vectors: int, index_key: str, current_memory_available: Optional[str], vec_dim: Optional[int]
+    nb_vectors: int, index_key: str, current_memory_available: Optional[str], vec_dim: Optional[int],
 ) -> int:
     """
     Function that determines the number of training points necessary to
@@ -73,12 +74,14 @@ def get_optimal_train_size(
 
     if matching:
         nb_clusters = index_key_to_nb_cluster(index_key)
-        points_per_cluster: float = 100
+        points_per_cluster: int = 100
 
         # compute best possible number of vectors to give to train the index
         # given memory constraints
         if current_memory_available and vec_dim:
-            memory_per_cluster_set = compute_memory_necessary_for_training(nb_clusters, index_key, vec_dim)
+            memory_per_cluster_set = compute_memory_necessary_for_training_wrapper(
+                points_per_cluster, index_key, vec_dim
+            )
             size = cast_memory_to_bytes(current_memory_available)
             points_per_cluster = max(min(size / memory_per_cluster_set, points_per_cluster), 31.0)
 
@@ -418,3 +421,85 @@ def get_optimal_hyperparameters(
     )
 
     return hyperparameter_str_from_param(optimal_param)
+
+
+def optimize_and_measure_index(
+    embedding_column_name: str,
+    embeddings_file_paths: List[str],
+    file_format: str,
+    index: faiss.Index,
+    index_infos_path: Optional[str],
+    index_key: str,
+    index_param: Optional[str],
+    index_path: Optional[str],
+    max_index_query_time_ms: float,
+    save_on_disk: bool,
+    use_gpu: bool,
+):
+    """Optimize one index by selecting the best hyperparameters and calculate its metrics"""
+    if index_param is None:
+        with Timeit(f"Computing best hyperparameters for index {index_path}", indent=1):
+            index_param = get_optimal_hyperparameters(index, index_key, max_speed_ms=max_index_query_time_ms)
+    # Set search hyperparameters for the index
+    set_search_hyperparameters(index, index_param, use_gpu)
+    logger.info(f"The best hyperparameters are: {index_param}")
+    metric_infos: Dict[str, Union[str, float, int]] = {"index_key": index_key, "index_param": index_param}
+    with Timeit("Compute fast metrics", indent=1):
+        metric_infos.update(
+            compute_fast_metrics(
+                embeddings_file_paths, index, file_format=file_format, embedding_column_name=embedding_column_name
+            )
+        )
+    if save_on_disk:
+        with Timeit("Saving the index on local disk", indent=1):
+            with fsspec.open(index_path, "wb").open() as f:
+                faiss.write_index(index, faiss.PyCallbackIOWriter(f.write))
+            with fsspec.open(index_infos_path, "w").open() as f:
+                json.dump(metric_infos, f)
+
+    return metric_infos
+
+
+def optimize_and_measure_indices(
+    indices_folder: str,
+    embedding_column_name: str,
+    embeddings_file_paths: List[str],
+    file_format: str,
+    index_infos_path: Optional[str],
+    index_key: str,
+    index_param: Optional[str],
+    index_path: Optional[str],
+    max_index_query_time_ms: float,
+    save_on_disk: bool,
+    use_gpu: bool,
+):
+    """Optimize a list of indices by selecting the best hyperparameters and calculate their metrics"""
+    fs, _ = fsspec.core.url_to_fs(indices_folder)
+    indices_file_paths = fs.ls(indices_folder, detail=False)
+    index_path2_metric_infos: Dict[str, Dict] = {}
+    suffix_width = int(math.log10(len(indices_file_paths))) + 1
+    with tempfile.TemporaryDirectory() as local_indices_folder:
+        parallel_download_indices_from_remote(
+            fs=fs, indices_file_paths=indices_file_paths, dst_folder=local_indices_folder
+        )
+        for i, tmp_filename in enumerate(sorted(os.listdir(local_indices_folder)), 1):
+            index_filepath = os.path.join(local_indices_folder, tmp_filename)
+            index = faiss.read_index(index_filepath)
+            if save_on_disk and index_path and index_infos_path:
+                cur_index_path = index_path + str(i).zfill(suffix_width)
+                cur_index_infos_path = index_infos_path + str(i)
+            metric_infos = optimize_and_measure_index(
+                embedding_column_name,
+                embeddings_file_paths,
+                file_format,
+                index,
+                cur_index_infos_path,
+                index_key,
+                index_param,
+                cur_index_path,
+                max_index_query_time_ms,
+                save_on_disk,
+                use_gpu,
+            )
+            index_path2_metric_infos[cur_index_path] = metric_infos
+    return index_path2_metric_infos
